--[[
Create visual elements.
Check items in DB against game.
Correct names, remove dead items.


DToX, 2013-2014
Licensed under the same terms as Lua itself.
--]]





-- Variables and constants
--

local checkStartTime;
local checkEndTime;
local IDs; -- Used for loop, contains all IDs
local checkIndex; -- Index for loop
local namesFromGame; -- Pairs "<ID>" = "<Name>"
local outdatedNames; -- Pairs "<ID>" = { "<NameFromDB>", "<NameFromGame>" }
local removedItems; -- Pairs "<ID>" = ""
local cntTotal;
local cntFound;
local cntSameNames;
local cntDifferentNames;
local LOOP_MAX_ITEMS = 100; -- Number of items to check in one loop
local REPORT_EVERY = 1000; -- Show status every N checked items (% operator)
-- Main window, carrier and offsets size
local MAIN_CARRIER_WIDTH = 450;
local MAIN_CARRIER_HEIGHT = 300;
local MAIN_CARRIER_OFFSET = 25;
local isCreated = false;


-- Forward declarations
--
local qs;
local btn_checkDB;
local btn_updateNames;
local btn_removeItems;
local txt_result;





-- Functions
--

-- Sets initial values, enumerates IDs from DB, gets total number of items in DB
local function checkDB()
   txt_result:SetText( "Checking..." );
   IDs = {};
   namesFromGame = {};
   outdatedNames = {};
   removedItems = {};
   checkIndex = 1;
   cntFound = 0;
   cntSameNames = 0;
   cntDifferentNames = 0;

   -- Enumerate IDs. Get total number of items
   local items = data.Items;
   for ID in pairs( items ) do
      local i = #IDs + 1;
      IDs[i] = ID;
   end
   cntTotal = #IDs;
   if ( cntTotal == 0 ) then
      txt_result:SetText( "Nothing to check. DB is empty." );
      return;
   end

   -- Call qs:Update
   qs:SetWantsUpdates( true );
end

-- Checks item. Returns item's information
local function checkItem( decID ) -- Decimal ID. Example: 1879049233
   qs:SetShortcut( nil );
   local ok, sc = pcall( addShortcut, qs, decID );
   if ( ok ) then
      local info = sc:GetItem():GetItemInfo();
      return true, info;
   else
      return false;
   end
end

-- Counts found and not found items
-- Gets names of items from game, IDs of not found items
-- Checks for difference in names
local function checkItems()
   local i = 1;
   local items = data.Items;
   local numID = IDs[checkIndex];
   while ( numID and i <= LOOP_MAX_ITEMS ) do
      -- Check ID
      local ok, info = checkItem( numID );
      local strID = tostring( numID );
      if ( ok ) then
         -- Increase counter, save name
         cntFound = cntFound + 1;
         local name = info:GetName();
         namesFromGame[numID] = name;

         -- Compare names, increase counters. If different, save ID and names
         local nameFromDB = items[numID][1][language];
         if ( nameFromDB ~= name ) then
            outdatedNames[strID] = {
               ["FromDB  "] = nameFromDB;
               ["FromGame"] = name;
            };
            cntDifferentNames = cntDifferentNames + 1;
         else
            cntSameNames = cntSameNames + 1;
         end
      else
         -- If not found, save ID
         removedItems[strID] = "";
      end

      -- Show progress and free memory
      if ( checkIndex % REPORT_EVERY == 0 ) then
         txt_result:SetText( "Checking...\nIDs checked: " .. checkIndex .. " (names found: " .. cntFound .. ")." );

         collectgarbage();
      end

      -- Increase loop's counter
      i = i + 1;
      checkIndex = checkIndex + 1;
      numID = IDs[checkIndex];

      -- No more items. Show status
      if ( not numID ) then
         -- Elapsed time
         checkEndTime = Turbine.Engine.GetLocalTime();
         local minutes = ( checkEndTime - checkStartTime ) / 60;
         minutes = string.format( "%.3f", minutes );
         txt_result:SetText( "Done. Elapsed minutes: " .. minutes .. "." );

         -- Status: items
         local cntNotFound = Table.len( removedItems );
         txt_result:AppendText( "\n[Items] Found + Not found = Sum / Total:\n"
         .. cntFound .. " + " .. cntNotFound .. " = " .. cntFound + cntNotFound .. " / " .. cntTotal .. "." );

         -- Status: names
         txt_result:AppendText( "\n[Names] Same + Different = Sum / Total (Total without Not found):\n"
         .. cntSameNames .. " + " .. cntDifferentNames .. " = " .. cntSameNames + cntDifferentNames .. " / " .. cntTotal .. " (" .. cntTotal - cntNotFound .. ")" );

         -- Update results
         data.NotFound = removedItems;
         data.OutdatedNames = outdatedNames;

         -- Enable buttons or show message
         if ( cntDifferentNames > 0 ) then
            btn_updateNames:SetEnabled( true );
         else
            txt_result:AppendText( "\n\nNothing to update. Button stays disabled." );
         end
         if ( cntNotFound > 0 ) then
            btn_removeItems:SetEnabled( true );
         else
            txt_result:AppendText( "\n\nNothing to remove. Button stays disabled." );
         end

         return;
      end
   end

   qs:SetWantsUpdates( true );
end

-- Updates names in DB with names from game or use current name
local function updateNames()
   local items = data.Items;
   btn_checkDB:SetEnabled( false );
   txt_result:SetText( "Updating names..." );
   for numID in pairs( items ) do
      items[numID][1][language] = namesFromGame[numID] or items[numID][1][language];
   end
   txt_result:AppendText( " Done." );
   btn_checkDB:SetEnabled( true );
   btn_updateNames:SetEnabled( false );
end

-- Remove dead items from DB
local function removeDeadItems()
   local items = data.Items;
   btn_checkDB:SetEnabled( false );
   txt_result:SetText( "Removing items..." );
   for strID in pairs( removedItems ) do
      local numID = tonumber( strID );
      items[numID] = nil;
   end
   txt_result:AppendText( " Done." );
   btn_checkDB:SetEnabled( true );
   btn_removeItems:SetEnabled( false );
end

-- Corrects main window's size
local function correctWindowSize( obj )
   local NEW_MAIN_WINDOW_WIDTH = MAIN_CARRIER_WIDTH + MAIN_CARRIER_OFFSET * 2;
   local b = getBorders( obj );
   local newMainWindowHeight = MAIN_CARRIER_HEIGHT + MAIN_CARRIER_OFFSET + b.Top;
   win_main:SetSize( NEW_MAIN_WINDOW_WIDTH, newMainWindowHeight );
end



-- Visuals
--

-- Fills carrier with visual elements or just toggles visibility
-- Returns some elements for manipulation
function showItems( mainCarrier )
   -- Main carrier for elements (offseted from window's borders)
   local mainCarrier = mainCarrier;

   correctWindowSize( mainCarrier );

   if ( isCreated ) then
      mainCarrier:SetVisible( true );
      return;
   end

   mainCarrier:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT );

   -- Button: Check DB
   btn_checkDB = newObject( mainCarrier, "Button", 130, "Check database" );
   btn_checkDB:SetPosition( 0, 0 );
   function btn_checkDB:MouseHover()
      Hint.show( self, "Scan game, collect data and show statistics (number of items, differences in names)." );
   end
   function btn_checkDB:MouseClick()
      checkStartTime = Turbine.Engine.GetLocalTime();
      checkDB();
   end

   -- Button: Update names
   btn_updateNames = newObject( mainCarrier, "Button", 130, "Update names" );
   btn_updateNames:SetEnabled( false );
   local b = getBorders( btn_checkDB );
   btn_updateNames:SetPosition( b.Left, b.Bottom + 10 );
   function btn_updateNames:MouseHover()
      Hint.show( self, "Replace names in DB with names from game.\nActual names can be found in *.plugindata file." );
   end
   function btn_updateNames:MouseClick()
      updateNames();
   end

   -- Button: Remove dead items
   btn_removeItems = newObject( mainCarrier, "Button", 130, "Remove items" );
   btn_removeItems:SetEnabled( false );
   local b = getBorders( btn_updateNames );
   btn_removeItems:SetPosition( b.Right + 10, b.Top );
   function btn_removeItems:MouseHover()
      Hint.show( self, "Remove dead items from DB.\nActual IDs can be found in *.plugindata file." );
   end
   function btn_removeItems:MouseClick()
      removeDeadItems();
   end

   -- Result: label
   local lbl_result = newObject( mainCarrier, "Label", 200, "Sanity check:" );
   local b = getBorders( btn_updateNames );
   lbl_result:SetPosition( b.Left, b.Bottom + 20 );

   -- Result: text
   txt_result = newObject( mainCarrier, "Text", 0, "" );
   txt_result:SetReadOnly( true );
   local b = getBorders( lbl_result );
   txt_result:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT - b.Bottom );
   txt_result:SetPosition( b.Left, b.Bottom );

   -- Quickslot for checking items
   qs = Turbine.UI.Lotro.Quickslot();
   function qs:Update()
      self:SetWantsUpdates( false );
      checkItems();
   end

   -- Set flag, that elements are created, and return some elements for manipulation
   isCreated = true;
   return txt_result, btn_updateNames, btn_removeItems;
end